(ns puppetlabs.activity.services
  (:require [clojure.tools.logging :as log]
            [puppetlabs.kitchensink.json :refer [add-common-json-encoders!]]
            [puppetlabs.trapperkeeper.services :refer [service-context]]
            [puppetlabs.trapperkeeper.core :refer [defservice]]
            [puppetlabs.activity.core :as activity]
            [puppetlabs.activity.http :as http]
            [puppetlabs.activity.postgres :as postgres]
            [compojure.core :refer [context]]
            [ring.util.servlet :refer [servlet]]))

(def fallback-db-spec {:subprotocol "postgresql"
                       :subname     "pe-activity"
                       :user        "pe-activity"
                       :password    "pe-activity"})

(defn- config->db-spec
  [{:keys [database]}]
  (merge fallback-db-spec database))

(defn- add-url-prefix
  "Puts the application in an optional url prefix. If none is given just
  returns the original handler."
  [prefix app]
  (if (seq prefix)
    (context prefix [] app)
    app))

(defprotocol ActivityService
  (shutdown [this] "Shut down the activity service."))

(defprotocol ActivityReportingService
  (report-activity! [this event-bundle] "Stores activity events."))

(defservice activity-reporting-service
  ActivityReportingService
  [[:ConfigService get-in-config]
   [:RbacService get-subject]]
  (init [this context]
        (let [db (-> (get-in-config [:activity])
                     config->db-spec
                     postgres/new-db)]
          (assoc context :db db)))
  (report-activity! [this event-bundle]
                    (activity/report-activity! (:db (service-context this)) get-subject event-bundle)))

(defservice activity-service
            ActivityService
            [[:ConfigService get-in-config]
             [:WebroutingService add-servlet-handler get-route]
             [:RbacAuthnMiddleware wrap-authenticated]
             [:RbacAuthzService is-permitted?]]

            (start [this context]
                   (let [config (get-in-config [:activity])
                         api-prefix (get-route this)
                         db-spec (config->db-spec config)
                         app-config (merge config {:db (postgres/new-db db-spec)
                                                   :login-path (get-in-config [:global :login-path])})
                         app (servlet (add-url-prefix api-prefix (http/app wrap-authenticated is-permitted? app-config)))]
                     (postgres/migrate! db-spec)
                     (add-common-json-encoders!)
                     (add-servlet-handler this app)
                     context))

            (shutdown [_] (log/info "Activity service shutting down.")))

(defservice initdb
            [[:ConfigService get-in-config]
             [:ShutdownService request-shutdown]]
            (start [_ context]
                   (let [config (get-in-config [:activity])
                         db-spec (config->db-spec config)]
                     (postgres/drop-public-tables! db-spec)
                     (postgres/drop-public-functions! db-spec)
                     (postgres/migrate! db-spec)
                     (request-shutdown)
                     context)))
